;;; dijkstra-gallenbacher.scm
(require (lib "breakpoint.ss"))
(load "dijskstra-gallenbacher-graph.scm")

; Anmerkungen:
; - markierte Orte sind bei Gallenbacher als rot bezeichnet.
; - In der Liste der markierten Kanten finden sich immer die Kanten zu dem Ort.
;   Zu dem Ort fuehrt aber jeweils genau eine Kante (im Gegensatz zu Kanten davon weg).
;   Daher kann man die Kante zu dem Ort aus der Liste der markierten Kanten entfernen,
;   indem man die Kante sucht, die an zweiter Position den Ort, also als Zielort hat.

;;; ===== alle Orte =====
(define
  alle-Orte
  '(A B C D E F G H I K L M N O P X Y Z))
(define anzahl-Orte (length alle-Orte))

;;; ===== kante-von-nach =====
; sucht die Kante mit ihrer Bewertung aus dem Graphen heraus.
(define
  (kante-von-nach von nach graph)
  (cond
    ((null? graph) (error "Eine Kante von " von " nach " nach "existiert nicht!"))
    ((not (equal? (first (first graph)) von)) (kante-von-nach von nach (rest graph)))
    ((not (equal? (second (first graph)) nach)) (kante-von-nach von nach (rest graph)))
    (else (first graph))))
; Test: (kante-von-nach 'I 'B gallenbacher-graph-kanten)

;;; ===== kante-nach =====
; sucht die Kante mit mit dem Zielort ihrer Bewertung aus markierteStrecken heraus.
(define
  (kante-nach nach markierteStrecken)
  (cond
    ((null? markierteStrecken) (error "Eine Kante nach " nach "existiert nicht!"))
    ((not (equal? (second (first markierteStrecken)) nach)) (kante-nach nach (rest markierteStrecken)))
    (else (first markierteStrecken))))

;;; ===== ort-mit-kleinster-kennung =====
; Der Ort mit minimalem Wert wird als erster Ort weitergereicht.
(define
  (ort-mit-kleinster-kennung hatKennung)
  (cond
    ((null? hatKennung) (error "Die Liste hatKennung darf nicht leer sein"))
    ((null? (rest hatKennung)) (first (first hatKennung)))
    ((< (second (first hatKennung)) (second (second hatKennung)))
     (ort-mit-kleinster-kennung (cons (first hatKennung) (rest (rest hatKennung)))))
    (else
     (ort-mit-kleinster-kennung (rest hatKennung)))))
; Test: (ort-mit-kleinster-kennung '((A 3)(B 1) (C 5)))

;;; ===== dijkstra =====
; aktuell: eine Stadt A ...
; nachbarn: Liste von Listen von Orten mit summe [ggf. neue Kennung]((A 23) ...)
; markierte: [rote] fertig eingebaute Orte mit ihren Bewertungen ((A 23) ...)
; hatKennung: Liste von Listen von Orten mit ihren Kennungen ((A 23) ...)
(define
  (dijkstra aktuell nachbarn markierte hatKennung ohneKennung markierteStrecken graph)
  (define
    (nichts-machen)
    (dijkstra aktuell (rest nachbarn) markierte hatKennung ohneKennung markierteStrecken graph))
  (define (erster-zielort) (second (first nachbarn)))
  (define (aktuelle-Kante) (kante-von-nach aktuell (erster-zielort) graph))
  ; - summe -
  (define (summe) (+ (second (assoc aktuell markierte)) (third (first nachbarn))))
  ; - kennung-hinzufuegen -
  (define (kennung-hinzufuegen hatKennung) (cons (list (erster-zielort) (summe)) hatKennung))
  ;;; ----- Begin der Hauptfunktion -----
  (cond
    ((= anzahl-Orte (length markierte))
     ; Lösung ist gefunden. Die Orte mit ihren Bewertungen und die aufgenommenen 
     ; Strecken werden ausgegeben.
     (list markierte markierteStrecken))
    ; Alle Nachbarn bearbeitet, aktuelle Stadt neu definieren
    ((null? nachbarn)
     (let
         ((aktuell (ort-mit-kleinster-kennung hatKennung)))
       (let
           ((nachbarn (kanten-von aktuell graph))
            (markierte (cons (assoc aktuell hatKennung) markierte))
            (hatKennung (remove (assoc aktuell hatKennung) hatKennung)))
         (dijkstra aktuell nachbarn markierte hatKennung ohneKennung markierteStrecken graph))))
    ; Beginn der Auswertungen der inneren Schleife:
    ((assoc (erster-zielort) markierte)
     ; ist "rot" markiert:  nichts machen.
     (nichts-machen))
    ((not (assoc (erster-zielort) hatKennung))
     ; hat noch keine Bewertung (Kennung): Kennung zuweisen und Strecke zur aktuellen Stadt markieren.
     (dijkstra
      aktuell
      (rest nachbarn)
      markierte
      (kennung-hinzufuegen hatKennung)
      (remove (erster-zielort) ohneKennung)
      (cons (aktuelle-Kante) markierteStrecken)
      graph))
    ((<= (second (assoc (erster-zielort) hatKennung)) (summe))
     ; Kennzahl kleiner oder gleich der Summe: nichts machen.
     (nichts-machen))
    (else
     ; Kennzahl ist größer:
     ; Kennzahl ersetzen, Strecke aus den markierten entfernen (s.o.)
     ; neue Kennung setzen und als markierte Strecke einfuegen.
     (dijkstra
      aktuell
      (rest nachbarn)
      markierte
      (kennung-hinzufuegen (remove (assoc (erster-zielort) hatKennung) hatKennung))
      ohneKennung
      (cons
       (aktuelle-Kante)
       (remove
        (kante-nach (erster-zielort) markierteStrecken)
        markierteStrecken))
      graph))))

;;; ===== Testaufruf =====´
; (dijkstra aktuell nachbarn markierte hatKennung ohneKennung strecken markierteStrecken graph)
(define
  (starte-dijkstra start graph)
  (dijkstra
   start
   (kanten-von start graph)
   (list (list start 0))
   '()
   (remove start alle-Orte)
   '()
   graph))

(starte-dijkstra 'I gallenbacher-graph-kanten)

;;; ===== kuerzester-weg =====
; Sucht aus der bestimmten Kantenliste den kürzesten Weg zu einem vorgegebenen Zielpunkt heraus.
; Es wird am Ziel begonnen und jeweils die Kante mit dem Ort an zweiter Position herausgesucht.
(define
  (kuerzester-weg von nach kantenliste)
  (define
    (schritt aktueller-knoten kante weg)
    (cond
      ((equal? von aktueller-knoten) weg)
      ((not kante)
       (error "Kante mit " aktueller-knoten "nicht enthalten"))
      (else
       (schritt
        (first kante)
        (finde-kante (first kante) kantenliste)
        (cons kante weg)))))
  (schritt
   nach
   (finde-kante nach kantenliste) '()))
;;; ----- finde-kante -----
; Hilfsfunktion zum finden der Kante mir dem Knoten an zweiter Position
(define
  (finde-kante knoten kantenliste)
  (cond
    ((null? kantenliste) #f)
    ((equal? knoten (second (first kantenliste)))
     (first kantenliste))
    (else
     (finde-kante knoten (rest kantenliste)))))
;;; ----- weg-laenge -----
; Hilfsfunktion zur Berechnung der Weglänge.
(define
  (weg-laenge weg)
  (apply + (map third weg)))

;;; ===== starte-suche =====
; Aufrufhülle für die Suche des kürzesten Weges
(define
  (starte-suche von nach graph)
  (kuerzester-weg von nach (second (starte-dijkstra von graph))))

; Testaufruf:
(starte-suche 'I 'O gallenbacher-graph-kanten)
(weg-laenge (starte-suche 'I 'O gallenbacher-graph-kanten))


